@@ -1,19 +1,19 @@
 use std::io::{self, Write};
 use std::path::PathBuf;
-use clap::{Command, Arg, ArgMatches};
+use clap::{Command, Arg};
 
 use sphinx::frontend;
 use sphinx::source::{ModuleSource, SourceText};
 use sphinx::parser::stmt::{StmtMeta, Stmt, StmtList, ControlFlow};
 use sphinx::parser::expr::Expr;
 use sphinx::parser::primary::Atom;
-use sphinx::parser::lvalue::{Declaration, DeclType, LValue};
+use sphinx::parser::lvalue::{LValue, AssignType, Assignment};
 use sphinx::codegen::{Program, CompiledProgram};
 use sphinx::runtime::{Module, VirtualMachine, Gc};
 use sphinx::runtime::module::GlobalEnv;
 use sphinx::runtime::strings::StringInterner;
 use sphinx::debug::symbol::resolver::BufferedResolver;
-use sphinx::stdlib;
+use sphinx::builtins;
 
 fn main() {
     env_logger::init();
@@ -39,11 +39,6 @@ fn main() {
             .short('i')
             .help("Drop into an interactive REPL after executing")
         )
-        .arg(
-            Arg::new("parse_only")
-            .short('P')
-            .help("Parse and print AST instead of executing")
-        )
         .arg(
             Arg::new("compile_only")
             .short('d')
@@ -59,31 +54,25 @@ fn main() {
     let args = app.get_matches();
     
     let source;
-    let name;
     if let Some(s) = args.value_of("cmd") {
         source = ModuleSource::String(s.to_string());
-        name = "<cmd>";
     } else if let Some(s) = args.value_of("file") {
         source = ModuleSource::File(PathBuf::from(s));
-        name = s;
     } else {
-        let repl_env = stdlib::create_prelude();
+        let repl_env = builtins::create_prelude();
         Repl::new(version.to_string(), repl_env).run();
         
         return;
     }
     
-    if args.is_present("parse_only") {
-        parse_and_print_ast(&args, name, &source);
-    }
-    else if args.is_present("compile_only") {
+    if args.is_present("compile_only") {
         unimplemented!()
     }
     else if args.is_present("interactive") {
         if let Some(build) = build_program(&source) {
             let program = Program::load(build.program);
             
-            let repl_env = stdlib::create_prelude();
+            let repl_env = builtins::create_prelude();
             let main_module = Module::with_env(Some(source), program.data, repl_env);
             
             let vm = VirtualMachine::new(main_module, &program.main);
@@ -99,7 +88,7 @@ fn main() {
     else if let Some(build) = build_program(&source) {
         let program = Program::load(build.program);
         
-        let main_env = stdlib::create_prelude();
+        let main_env = builtins::create_prelude();
         let main_module = Module::with_env(Some(source), program.data, main_env);
         
         let vm = VirtualMachine::new(main_module, &program.main);
@@ -143,29 +132,6 @@ fn run_debugger(vm: VirtualMachine) {
     println!("Execution stopped.");
 }
 
-fn parse_and_print_ast(_args: &ArgMatches, name: &str, source: &ModuleSource) {
-    let source_text = match source.read_text() {
-        Ok(source_text) => source_text,
-        
-        Err(error) => {
-            println!("Error reading source: {}.", error);
-            return;
-        },
-    };
-    
-    let mut interner = StringInterner::new();
-    let parse_result = sphinx::parse_source(&mut interner, source_text);
-    
-    match parse_result {
-        Err(errors) => {
-            println!("Errors in file \"{}\":\n", name);
-            frontend::print_source_errors(source, &errors);
-        },
-        Ok(ast) => println!("{:#?}", ast),
-    }
-}
-
-
 //////// REPL ////////
 const PROMT_START: &str = ">>> ";
 const PROMT_CONTINUE: &str = "... ";
@@ -319,10 +285,11 @@ impl Repl {
         
         // bind the result expression to a global name
         let result_name = interner.get_or_intern("_");
-        let result_decl = Expr::Declaration(Box::new(Declaration {
-            decl: DeclType::Immutable,
+        let result_decl = Expr::Assignment(Box::new(Assignment {
+            assign: AssignType::DeclImmutable,
             lhs: LValue::Identifier(result_name),
-            init: result_expr,
+            rhs: result_expr,
+            op: None,
         }));
         ast.push(StmtMeta::new(Stmt::Expression(result_decl), symbol));
         
