@@ -2,14 +2,15 @@
 
 use core::iter;
 
-use crate::language::{IntType, FloatType, InternSymbol};
+use crate::language::{IntType, FloatType, InternSymbol, Access};
 use crate::parser::stmt::{StmtMeta, Stmt, Label, StmtList, ControlFlow};
 use crate::parser::expr::{Expr, ExprMeta, ExprBlock, ConditionalBranch};
 use crate::parser::primary::{Atom, Primary, AccessItem};
-use crate::parser::lvalue::{Assignment, Declaration, LValue, DeclType};
+use crate::parser::lvalue::{LValue, AssignType};
 use crate::parser::fundefs::{FunctionDef, SignatureDef};
 use crate::parser::operator::{UnaryOp, BinaryOp};
 use crate::runtime::strings::{StringInterner};
+use crate::runtime::errors::ErrorKind;
 use crate::debug::symbol::{DebugSymbol, ChunkSymbols, DebugSymbolTable};
 
 mod scope;
@@ -24,7 +25,7 @@ pub use opcodes::{OpCode, LocalIndex};
 pub use chunk::{UnloadedProgram, Program, ProgramData, Chunk};
 pub use consts::{ConstID, Constant};
 pub use funproto::{FunctionID, FunctionProto, UpvalueTarget};
-pub use errors::{CompileResult, CompileError, ErrorKind};
+pub use errors::{CompileResult, CompileError};
 
 use scope::{ScopeTracker, ScopeTag, Scope, LocalName, InsertLocal, ControlFlowTarget};
 use chunk::{ChunkBuilder, ChunkInfo, ChunkBuf};
@@ -33,41 +34,6 @@ use funproto::{UnloadedFunction, UnloadedSignature, UnloadedParam};
 
 // Helpers
 
-struct AssignmentRef<'a> {
-    lhs: &'a LValue,
-    op: Option<BinaryOp>,
-    rhs: &'a Expr,
-    nonlocal: bool,
-}
-
-impl<'a> From<&'a Assignment> for AssignmentRef<'a> {
-    fn from(assign: &'a Assignment) -> Self {
-        Self {
-            lhs: &assign.lhs,
-            op: assign.op,
-            rhs: &assign.rhs,
-            nonlocal: assign.nonlocal,
-        }
-    }
-}
-
-struct DeclarationRef<'a> {
-    decl: DeclType,
-    lhs: &'a LValue,
-    init: &'a Expr,
-}
-
-impl<'a> From<&'a Declaration> for DeclarationRef<'a> {
-    fn from(decl: &'a Declaration) -> Self {
-        Self {
-            decl: decl.decl,
-            lhs: &decl.lhs,
-            init: &decl.init,
-        }
-    }
-}
-
-
 #[derive(Clone, Copy, PartialEq, Eq, Hash)]
 enum JumpOffset {
     Short(i16),
@@ -328,6 +294,17 @@ impl CodeGenerator<'_> {
         Ok(())
     }
     
+    fn emit_load_error(&mut self, symbol: Option<&DebugSymbol>, error: ErrorKind, message: &str) -> CompileResult<()> {
+        let cid = self.builder_mut().get_or_insert_error(error, message)?;
+        
+        if cid <= u8::MAX.into() {
+            self.emit_instr_byte(symbol, OpCode::LoadConst, u8::try_from(cid).unwrap());
+        } else {
+            self.emit_instr_data(symbol, OpCode::LoadConst16, &cid.to_le_bytes());
+        }
+        Ok(())
+    }
+    
     fn make_function(&mut self, function: UnloadedFunction) {
         self.builder_mut().insert_function(function)
     }
@@ -358,7 +335,7 @@ impl CodeGenerator<'_> {
             
             // if we *still* don't have the right width, just abort
             if new_width != new_opcode.instr_len() {
-                return Err(ErrorKind::InternalLimit("could not calculate jump offset").into());
+                return Err("could not calculate jump offset".into());
             }
             
             jump_offset = new_offset;
@@ -400,7 +377,7 @@ impl CodeGenerator<'_> {
             
             // if we *still* don't have the right width, just abort
             if new_width != new_opcode.instr_len() {
-                return Err(ErrorKind::InternalLimit("could not calculate jump offset").into());
+                return Err("could not calculate jump offset".into());
             }
             
             jump_offset = new_offset;
@@ -429,7 +406,7 @@ impl CodeGenerator<'_> {
             return Ok(JumpOffset::Long(offset));
         }
         
-        Err(ErrorKind::InternalLimit("could not calculate jump offset").into())
+        Err("could not calculate jump offset".into())
     }
 }
 
@@ -639,7 +616,12 @@ impl CodeGenerator<'_> {
         // find the target scope
         let target_depth = match self.scopes().resolve_control_flow(ControlFlowTarget::Break(label.copied())) {
             Some(scope) => scope.depth(),
-            None => return Err(ErrorKind::CantResolveBreak(label.copied()).into()),
+            None => {
+                let message =
+                    if label.is_some() { "can't find loop or block with matching label for \"break\"" }
+                    else { "\"break\" outside of loop or block" };
+                return Err(message.into());
+            },
         };
         
         // drop all scopes up to and including the target
@@ -670,7 +652,7 @@ impl CodeGenerator<'_> {
                 self.emit_instr(symbol, OpCode::Nil);
             }
         } else if expr.is_some() {
-            return Err(CompileError::new(ErrorKind::InvalidBreakWithValue))
+            return Err("\"break\" with value outside of block expression".into())
         }
         
         // emit jump site, register with scope
@@ -689,7 +671,12 @@ impl CodeGenerator<'_> {
         // find the target scope
         let (target_depth, continue_target) = match self.scopes().resolve_control_flow(ControlFlowTarget::Continue(label.copied())) {
             Some(scope) => (scope.depth(), scope.continue_target()),
-            None => return Err(ErrorKind::CantResolveContinue(label.copied()).into()),
+            None => {
+                let message =
+                    if label.is_some() { "can't find loop with matching label for \"continue\"" }
+                    else { "\"continue\" outside of loop" };
+                return Err(message.into());
+            }
         };
         
         // drop all scopes up to and including the target
@@ -773,22 +760,25 @@ impl CodeGenerator<'_> {
             
             Expr::Primary(primary) => self.compile_primary(symbol, primary)?,
             
-            Expr::UnaryOp(op, expr) => {
-                self.compile_expr(symbol, expr)?;
-                self.emit_unary_op(symbol, op);
-            },
+            Expr::UnaryOp(op, expr) => self.compile_unary_op(symbol, *op, expr)?,
             
             Expr::BinaryOp(op, exprs) => {
-                let (ref lhs, ref rhs) = **exprs;
-                self.compile_expr(symbol, lhs)?;
-                self.compile_expr(symbol, rhs)?;
-                self.emit_binary_op(symbol, op);
+                let (lhs, rhs) = &**exprs;
+                self.compile_binary_op(symbol, *op, lhs, rhs)?;
             },
             
-            Expr::Declaration(decl) => self.compile_declaration(symbol, (&**decl).into())?,
-            Expr::Assignment(assign) => self.compile_assignment(symbol, (&**assign).into())?,
+            Expr::Assignment(assign) => {
+                if let Some(op) = assign.op {
+                    self.compile_update_assignment(symbol, op, assign.assign, &assign.lhs, &assign.rhs)?;
+                } else {
+                    self.compile_expr(symbol, &assign.rhs)?;
+                    self.compile_assignment(symbol, assign.assign, &assign.lhs)?;
+                }
+            },
             
-            Expr::Tuple(expr_list) => self.compile_tuple(symbol, expr_list)?,
+            Expr::Tuple(items) => self.compile_tuple(symbol, items)?,
+            
+            Expr::Ellipsis(..) => return Err("\"...\" is not allowed here".into()),
             
             Expr::Block { label, suite } => self.compile_block_expression(symbol, label.as_ref(), suite)?,
             Expr::IfExpr { branches, else_clause } => self.compile_if_expression(symbol, branches, else_clause.as_ref().map(|expr| &**expr))?,
@@ -805,7 +795,7 @@ impl CodeGenerator<'_> {
     
     fn compile_tuple(&mut self, symbol: Option<&DebugSymbol>, expr_list: &[ExprMeta]) -> CompileResult<()> {
         let len = u8::try_from(expr_list.len())
-            .map_err(|_| ErrorKind::InternalLimit("tuple length limit exceeded"))?;
+            .map_err(|_| "tuple length limit exceeded")?;
         
         for expr in expr_list.iter() {
             self.compile_expr_with_symbol(expr)?;
@@ -831,7 +821,13 @@ impl CodeGenerator<'_> {
             // Atom::Self_ => unimplemented!(),
             // Atom::Super => unimplemented!(),
             
-            Atom::Group(expr) => self.compile_expr(symbol, expr)?,
+            Atom::Group { modifier, inner } => {
+                // modifiers are not allowed outside of assignment
+                if let Some(modifier) = modifier {
+                    return Err("assignment modifier is not allowed here".into())
+                }
+                self.compile_expr(symbol, inner)?
+            },
         }
         Ok(())
     }
@@ -879,60 +875,97 @@ impl CodeGenerator<'_> {
             match item {
                 AccessItem::Attribute(name) => unimplemented!(),
                 AccessItem::Index(index) => unimplemented!(),
-                AccessItem::Invoke { args, unpack } => self.compile_invocation(symbol, args, unpack.as_ref())?,
+                AccessItem::Invoke(args) => self.compile_invocation(symbol, args)?,
             }
         }
         
         Ok(())
     }
     
-    fn compile_invocation(&mut self, symbol: Option<&DebugSymbol>, args: &[ExprMeta], unpack: Option<&ExprMeta>) -> CompileResult<()> {
+    fn compile_invocation(&mut self, symbol: Option<&DebugSymbol>, mut args: &[ExprMeta]) -> CompileResult<()> {
         // prepare argument list:
         // [ callobj arg[n] arg[0] ... arg[n-1] nargs ] => [ ret_value ] 
 
         let arg_len;
+        let mut unpack = None;
         
-        if let Some((arg_last, args)) = args.split_last() {
+        // process argument unpacking
+        if let Some((last, rest)) = args.split_last() {
+            let (expr, symbol) = last.clone().take();
+            match expr {
+                Expr::Ellipsis(None) => return Err("need a value to unpack".into()),
+                
+                Expr::Ellipsis(Some(inner)) => {
+                    args = rest;
+                    unpack.replace(ExprMeta::new(*inner, symbol));
+                }
+                
+                _ => { }
+            }
+            
+            if args.iter().any(|arg| matches!(arg.variant(), Expr::Ellipsis(..))) {
+                return Err("\"...\" can only be used to unpack the last argument".into());
+            }
+        }
+        
+        // compile arguments
+        if let Some((arg_last, args_rest)) = args.split_last() {
             
             self.compile_expr_with_symbol(arg_last)?;
-            
-            for arg_expr in args.iter() {
+            for arg_expr in args_rest.iter() {
                 self.compile_expr_with_symbol(arg_expr)?;
             }
             
-            arg_len = u8::try_from(args.len() + 1)
-                .map_err(|_| ErrorKind::InternalLimit("argument count limit exceeded"))?;
+            arg_len = u8::try_from(args.len())
+                .map_err(|_| "argument count limit exceeded")?;
             
         } else {
             
             arg_len = 0;
         }
         
+        self.emit_instr_byte(symbol, OpCode::UInt8, arg_len);
+        
         if let Some(seq_expr) = unpack {
-            self.compile_expr_with_symbol(seq_expr)?;
-            self.emit_instr_byte(symbol, OpCode::UInt8, arg_len);
+            self.compile_expr_with_symbol(&seq_expr)?;
             self.emit_instr(symbol, OpCode::CallUnpack);
         } else {
-            self.emit_instr_byte(symbol, OpCode::UInt8, arg_len);
             self.emit_instr(symbol, OpCode::Call);
         }
 
         Ok(())
     }
     
-    fn emit_unary_op(&mut self, symbol: Option<&DebugSymbol>, op: &UnaryOp) {
+    fn compile_unary_op(&mut self, symbol: Option<&DebugSymbol>, op: UnaryOp, expr: &Expr) -> CompileResult<()> {
+        self.compile_expr(symbol, expr)?;
         match op {
             UnaryOp::Neg => self.emit_instr(symbol, OpCode::Neg),
             UnaryOp::Pos => self.emit_instr(symbol, OpCode::Pos),
             UnaryOp::Inv => self.emit_instr(symbol, OpCode::Inv),
             UnaryOp::Not => self.emit_instr(symbol, OpCode::Not),
-        }
+        };
+        Ok(())
     }
     
-    fn emit_binary_op(&mut self, symbol: Option<&DebugSymbol>, op: &BinaryOp) {
+    fn compile_binary_op(&mut self, symbol: Option<&DebugSymbol>, op: BinaryOp, lhs: &Expr, rhs: &Expr) -> CompileResult<()> {
+        
+        if matches!(op, BinaryOp::And) {
+            return self.compile_shortcircuit_and(symbol, lhs, rhs);
+        }
+        if matches!(op, BinaryOp::Or) {
+            return self.compile_shortcircuit_or(symbol, lhs, rhs);
+        }
+        
+        self.compile_expr(symbol, lhs)?;
+        self.compile_expr(symbol, rhs)?;
+        self.emit_binary_op(symbol, op);
+        
+        Ok(())
+    }
+    
+    fn emit_binary_op(&mut self, symbol: Option<&DebugSymbol>, op: BinaryOp) {
         match op {
-            BinaryOp::And => unimplemented!(),
-            BinaryOp::Or => unimplemented!(),
+            BinaryOp::And | BinaryOp::Or => unreachable!(),
             
             BinaryOp::Mul => self.emit_instr(symbol, OpCode::Mul),
             BinaryOp::Div => self.emit_instr(symbol, OpCode::Div),
@@ -953,62 +986,99 @@ impl CodeGenerator<'_> {
             BinaryOp::GE => self.emit_instr(symbol, OpCode::GE),
             BinaryOp::EQ => self.emit_instr(symbol, OpCode::EQ),
             BinaryOp::NE => self.emit_instr(symbol, OpCode::NE),
-        }
+        };
     }
 }
 
 ///////// Declarations and Assignments /////////
 impl CodeGenerator<'_> {
-    fn compile_declaration(&mut self, symbol: Option<&DebugSymbol>, decl: DeclarationRef) -> CompileResult<()> {
-        match &decl.lhs {
-            LValue::Identifier(name) => if self.scopes().is_global_scope() {
-                self.compile_decl_global_name(symbol, decl.decl, *name, decl.init)
-            } else {
-                self.compile_decl_local_name(symbol, decl.decl, *name, decl.init)
+
+    
+    fn compile_update_assignment(&mut self, symbol: Option<&DebugSymbol>, op: BinaryOp, assign: AssignType, lhs: &LValue, rhs: &Expr) -> CompileResult<()> {
+        
+        let local_only = match assign {
+            AssignType::AssignLocal => true,
+            AssignType::AssignNonLocal => false,
+            
+            AssignType::DeclImmutable | AssignType::DeclMutable
+                => return Err("update-assignment is invalid when declaring a variable".into()),
+        };
+        
+        // TODO suport Attribute and Index LValues as well
+        match lhs {
+            LValue::Identifier(name) => {
+                self.compile_name_lookup(symbol, name)?;
+                self.compile_expr(symbol, rhs)?;
+                self.emit_binary_op(symbol, op);
+                
+                self.compile_assign_identifier(symbol, name, local_only)
             },
             
             LValue::Attribute(target) => unimplemented!(),
+            
             LValue::Index(target) => unimplemented!(),
             
-            LValue::Tuple(target_list) => match &decl.init {
-                
-                Expr::Tuple(init_list) => {
-                    if target_list.len() != init_list.len() {
-                        return Err(ErrorKind::TupleLenMismatch.into());
-                    }
-                    
-                    for (inner_lhs, inner_expr) in target_list.iter().zip(init_list.iter()) {
-                        let inner_symbol = inner_expr.debug_symbol();
-                        let inner_init = inner_expr.variant();
-                        
-                        let inner_decl = DeclarationRef {
-                            decl: decl.decl,
-                            lhs: inner_lhs,
-                            init: inner_init,
-                        };
-                        
-                        self.compile_declaration(Some(inner_symbol), inner_decl)
-                            .map_err(|err| err.with_symbol(*inner_symbol))?;
-                    }
-                    
-                    // declarations are also expressions
-                    let tuple_len = u8::try_from(init_list.len())
-                        .map_err(|_| ErrorKind::InternalLimit("max tuple length exceeded"))?;
-                    self.emit_instr_byte(symbol, OpCode::Tuple, tuple_len);
-                    
-                    Ok(())
-                },
-                
-                _ => unimplemented!(), // dynamic declaration
-            },
+            LValue::Tuple {..} | LValue::PackItem(..)
+                => Err("can't update-assign to this".into()),
+            
+            LValue::Modifier {..} => unreachable!(),
         }
     }
     
-    fn compile_decl_local_name(&mut self, symbol: Option<&DebugSymbol>, decl: DeclType, name: InternSymbol, init: &Expr) -> CompileResult<()> {
+    fn compile_assignment(&mut self, symbol: Option<&DebugSymbol>, mut assign: AssignType, mut lhs: &LValue) -> CompileResult<()> {
         
-        self.compile_expr(symbol, init)?;  // make sure to evaluate initializer first in order for shadowing to work
+        while let LValue::Modifier { modifier, lvalue } = lhs {
+            assign = *modifier;
+            lhs = lvalue;
+        }
         
-        match self.scopes_mut().insert_local(decl, LocalName::Symbol(name))? {
+        match lhs {
+            LValue::Tuple(items) => self.compile_assign_tuple(symbol, assign, items),
+            
+            LValue::PackItem(..) => {
+                let item = std::slice::from_ref(lhs);
+                self.compile_assign_tuple(symbol, assign, item)
+            }
+            
+            lhs => {
+                match assign {
+                    AssignType::AssignLocal => self.compile_assign_variable(symbol, lhs, false),
+                    AssignType::AssignNonLocal => self.compile_assign_variable(symbol, lhs, true),
+                    AssignType::DeclImmutable => self.compile_decl_variable(symbol, Access::ReadOnly, lhs),
+                    AssignType::DeclMutable => self.compile_decl_variable(symbol, Access::ReadWrite, lhs),
+                }
+            }
+        }
+    }
+
+    fn compile_decl_variable(&mut self, symbol: Option<&DebugSymbol>, access: Access, lhs: &LValue) -> CompileResult<()> {
+        match lhs {
+            LValue::Identifier(name) => if self.scopes().is_global_scope() {
+                self.compile_decl_global_name(symbol, access, *name)
+            } else {
+                self.compile_decl_local_name(symbol, access, *name)
+            },
+            
+            LValue::Tuple {..} => unreachable!(),
+            LValue::Modifier {..} => unreachable!(),
+            
+            _ => Err("not a variable name".into()),
+        }
+    }
+    
+    fn compile_decl_global_name(&mut self, symbol: Option<&DebugSymbol>, access: Access, name: InternSymbol) -> CompileResult<()> {
+        
+        self.emit_load_const(symbol, Constant::from(name))?;
+        match access {
+            Access::ReadOnly => self.emit_instr(symbol, OpCode::InsertGlobal),
+            Access::ReadWrite => self.emit_instr(symbol, OpCode::InsertGlobalMut),
+        }
+        Ok(())
+    }
+    
+    fn compile_decl_local_name(&mut self, symbol: Option<&DebugSymbol>, access: Access, name: InternSymbol) -> CompileResult<()> {
+        
+        match self.scopes_mut().insert_local(access, LocalName::Symbol(name))? {
             InsertLocal::CreateNew => 
                 self.emit_instr(symbol, OpCode::InsertLocal),
             
@@ -1019,81 +1089,27 @@ impl CodeGenerator<'_> {
         Ok(())
     }
     
-    fn compile_decl_global_name(&mut self, symbol: Option<&DebugSymbol>, decl: DeclType, name: InternSymbol, init: &Expr) -> CompileResult<()> {
-        
-        self.compile_expr(symbol, init)?;
-        
-        self.emit_load_const(symbol, Constant::from(name))?;
-        match decl {
-            DeclType::Immutable => self.emit_instr(symbol, OpCode::InsertGlobal),
-            DeclType::Mutable => self.emit_instr(symbol, OpCode::InsertGlobalMut),
-        }
-        Ok(())
-    }
+    // fn compile_decl_tuple(&mut self, symbol: Option<&DebugSymbol>, access: Access, targets: &[LValue]) -> CompileResult<()> {
+    //     self.compile_tuple_unpack(
+    //         symbol, targets, 
+    //         |self_, target| self_.compile_declaration(symbol, access, target)
+    //     )
+    // }
     
-    fn compile_assignment(&mut self, symbol: Option<&DebugSymbol>, assign: AssignmentRef) -> CompileResult<()> {
-        match assign.lhs {
-            // LValue::Identifier(name) if self.state.is_global_scope() => self.compile_assign_global_name(symbol, op, *name, &rhs),
-            LValue::Identifier(name) => self.compile_assign_identifier(symbol, name, assign),
+    fn compile_assign_variable(&mut self, symbol: Option<&DebugSymbol>, lhs: &LValue, allow_nonlocal: bool) -> CompileResult<()> {
+        
+        match lhs {
+            LValue::Identifier(name) => self.compile_assign_identifier(symbol, name, allow_nonlocal),
             
             LValue::Attribute(target) => unimplemented!(),
+            
             LValue::Index(target) => unimplemented!(),
             
-            LValue::Tuple(..) if assign.op.is_some() => {
-                Err(ErrorKind::CantUpdateAssignTuple.into())
-            },
-            
-            LValue::Tuple(target_list) => match assign.rhs {
-                
-                Expr::Tuple(rhs_list) => {
-                    if target_list.len() != rhs_list.len() {
-                        return Err(ErrorKind::TupleLenMismatch.into())
-                    }
-                    
-                    for (inner_lhs, inner_expr) in target_list.iter().zip(rhs_list.iter()) {
-                        let inner_symbol = inner_expr.debug_symbol();
-                        let inner_rhs = inner_expr.variant();
-                        
-                        let inner_assign = AssignmentRef {
-                            lhs: inner_lhs,
-                            op: None,
-                            rhs: inner_rhs,
-                            nonlocal: assign.nonlocal,
-                        };
-                        
-                        self.compile_assignment(Some(inner_symbol), inner_assign)
-                            .map_err(|err| err.with_symbol(*inner_symbol))?;
-                    }
-                    
-                    // assignments are also expressions
-                    let tuple_len = u8::try_from(rhs_list.len())
-                        .map_err(|_| ErrorKind::InternalLimit("max tuple length exceeded"))?;
-                    self.emit_instr_byte(symbol, OpCode::Tuple, tuple_len);
-                    
-                    Ok(())
-                },
-                
-                _ => unimplemented!(), // dynamic declaration
-                
-            },
+            _ => panic!("invalid assignment target"),
         }
     }
     
-    fn compile_assign_identifier(&mut self, symbol: Option<&DebugSymbol>, name: &InternSymbol, assign: AssignmentRef) -> CompileResult<()> {
-        
-        // Compile RHS
-        
-        if let Some(op) = assign.op {
-            // update assignment
-            self.compile_name_lookup(symbol, name)?;
-            self.compile_expr(symbol, assign.rhs)?;
-            self.emit_binary_op(symbol, &op);
-            
-        } else {
-            // normal assignment
-            self.compile_expr(symbol, assign.rhs)?;
-            
-        }
+    fn compile_assign_identifier(&mut self, symbol: Option<&DebugSymbol>, name: &InternSymbol, allow_nonlocal: bool) -> CompileResult<()> {
         
         // Generate assignment
         
@@ -1105,8 +1121,8 @@ impl CodeGenerator<'_> {
             let result = self.scopes().resolve_local(&local_name);
             
             if let Some(local) = result.cloned() {
-                if local.decl() != DeclType::Mutable {
-                    return Err(CompileError::from(ErrorKind::CantAssignImmutable));
+                if !local.mode().can_write() {
+                    return Err("can't assign to immutable local variable".into());
                 }
                 
                 self.emit_assign_local(symbol, local.index());
@@ -1115,15 +1131,15 @@ impl CodeGenerator<'_> {
             }
             
             // nonlocal keyword is not required in the global frame
-            if !assign.nonlocal && !self.scopes().is_global_frame() {
-                return Err(CompileError::from(ErrorKind::CantAssignNonLocal));
+            if !allow_nonlocal && !self.scopes().is_global_frame() {
+                return Err("can't assign to a non-local variable without the \"nonlocal\" keyword".into());
             }
             
             // check if an upvalue is found or can be created...
             if !self.scopes().is_global_frame() {
                 if let Some(upval) = self.scopes_mut().resolve_or_create_upval(&local_name)? {
-                    if upval.decl() != DeclType::Mutable {
-                        return Err(CompileError::from(ErrorKind::CantAssignImmutable));
+                    if !upval.mode().can_write() {
+                        return Err("can't assign to immutable local variable".into());
                     }
                     
                     let index = upval.index();
@@ -1151,6 +1167,88 @@ impl CodeGenerator<'_> {
             self.emit_instr_data(symbol, OpCode::StoreLocal16, &offset.to_le_bytes());
         }
     }
+    
+    fn compile_assign_tuple(&mut self, symbol: Option<&DebugSymbol>, assign: AssignType, mut item_targets: &[LValue]) -> CompileResult<()> {
+        
+        let mut error_jump_sites = Vec::new();
+        
+        self.emit_instr(symbol, OpCode::IterInit);  // iterate to yield values
+        
+        // process tuple packing assingment
+        let mut pack_target = None;
+        
+        if let Some((LValue::PackItem(pack_item), rest)) = item_targets.split_last() {
+            pack_target.replace(pack_item);
+            item_targets = rest;
+        }
+        
+        // compile to unrolled iteration
+        for (idx, target) in item_targets.iter().enumerate() {
+            
+            if matches!(target, LValue::PackItem(..)) {
+                return Err("\"...\" can only be used with the last item in tuple assignment".into());
+            }
+            
+            // check if there is an item left for this target
+            let error_jump = self.emit_dummy_jump(symbol, Jump::IfFalse);
+            error_jump_sites.push(error_jump);
+            
+            // advance the iterator and put the item on the stack
+            self.emit_instr(symbol, OpCode::IterNext);
+            
+            self.compile_assignment(symbol, assign, target)?;
+            self.emit_instr(symbol, OpCode::Pop);
+        }
+        
+        let done_jump_site;
+        let pop_after;
+        if let Some(Some(pack_target)) = pack_target {
+            // exhaust the rest of the iterator and assign to pack_target
+            self.emit_instr(symbol, OpCode::IterUnpack);
+            self.emit_instr(symbol, OpCode::TupleN);
+            
+            self.compile_assignment(symbol, assign, pack_target)?;
+            self.emit_instr(symbol, OpCode::Pop);
+            
+            done_jump_site = self.emit_dummy_jump(symbol, Jump::Uncond);
+            pop_after = false;
+        
+        } else if let Some(None) = pack_target {
+            // just discard the iterator
+            self.emit_instr_byte(symbol, OpCode::Drop, 2);
+            done_jump_site = self.emit_dummy_jump(symbol, Jump::Uncond);
+            pop_after = false;
+            
+        } else {
+            // if the iterator is finished we've succeeded
+            done_jump_site = self.emit_dummy_jump(symbol, Jump::PopIfFalse);
+            pop_after = true; // pop the iterator
+            
+            // too many items
+            let message = format!("too many values to unpack (expected {})", item_targets.len());
+            self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
+            self.emit_instr(symbol, OpCode::Error);
+        }
+        
+        // not enough items
+        let error_target = self.current_offset();
+        for jump_site in error_jump_sites.iter() {
+            self.patch_jump_instr(jump_site, error_target)?;
+        }
+        
+        let message = format!("not enough values to unpack (expected {})", item_targets.len());
+        self.emit_load_error(symbol, ErrorKind::UnpackError, message.as_str())?;
+        self.emit_instr(symbol, OpCode::Error);
+        
+        // cleanup
+        self.patch_jump_instr(&done_jump_site, self.current_offset())?;
+        
+        if pop_after {
+            self.emit_instr(symbol, OpCode::Pop);
+        }
+        
+        Ok(())
+    }
 }
 
 ///////// Blocks and If-Expressions /////////
@@ -1223,6 +1321,32 @@ impl CodeGenerator<'_> {
         
         Ok(())
     }
+    
+    fn compile_shortcircuit_and(&mut self, symbol: Option<&DebugSymbol>, lhs: &Expr, rhs: &Expr) -> CompileResult<()> {
+        self.compile_expr(symbol, lhs)?;
+        
+        let shortcircuit = self.emit_dummy_jump(symbol, Jump::IfFalse);
+        
+        self.emit_instr(symbol, OpCode::Pop);
+        self.compile_expr(symbol, rhs)?;
+        
+        self.patch_jump_instr(&shortcircuit, self.current_offset())?;
+        
+        Ok(())
+    }
+    
+    fn compile_shortcircuit_or(&mut self, symbol: Option<&DebugSymbol>, lhs: &Expr, rhs: &Expr) -> CompileResult<()> {
+        self.compile_expr(symbol, lhs)?;
+        
+        let shortcircuit = self.emit_dummy_jump(symbol, Jump::IfTrue);
+        
+        self.emit_instr(symbol, OpCode::Pop);
+        self.compile_expr(symbol, rhs)?;
+        
+        self.patch_jump_instr(&shortcircuit, self.current_offset())?;
+        
+        Ok(())
+    }
 }
 
 ///////// Function Definitions /////////
@@ -1246,8 +1370,8 @@ impl CodeGenerator<'_> {
         chunk_gen.scopes_mut().push_frame(symbol);
         
         // don't need to generate IN_LOCAL instructions for these, the VM should include them automatically
-        chunk_gen.scopes_mut().insert_local(DeclType::Immutable, LocalName::Receiver)?;
-        chunk_gen.scopes_mut().insert_local(DeclType::Immutable, LocalName::NArgs)?;
+        chunk_gen.scopes_mut().insert_local(Access::ReadOnly, LocalName::Receiver)?;
+        chunk_gen.scopes_mut().insert_local(Access::ReadOnly, LocalName::NArgs)?;
         
         // prepare argument list
         chunk_gen.compile_function_preamble(symbol, fundef)?;
@@ -1299,7 +1423,7 @@ impl CodeGenerator<'_> {
         let signature = &fundef.signature;
         
         for param in signature.required.iter() {
-            self.scopes_mut().insert_local(param.decl, LocalName::Symbol(param.name))?;
+            self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
             //self.emit_instr(None, OpCode::InsertLocal);
         }
         
@@ -1308,7 +1432,7 @@ impl CodeGenerator<'_> {
             self.compile_default_args(signature)?;
             
             for param in signature.default.iter() {
-                self.scopes_mut().insert_local(param.decl, LocalName::Symbol(param.name))?;
+                self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
                 // self.emit_instr(symbol, OpCode::InsertLocal);
             }
         }
@@ -1316,7 +1440,7 @@ impl CodeGenerator<'_> {
         if let Some(param) = &signature.variadic {
             self.compile_variadic_arg(signature)?;
 
-            self.scopes_mut().insert_local(param.decl, LocalName::Symbol(param.name))?;
+            self.scopes_mut().insert_local(param.mode, LocalName::Symbol(param.name))?;
             // self.emit_instr(symbol, OpCode::InsertLocal);
         }
 
@@ -1331,9 +1455,9 @@ impl CodeGenerator<'_> {
         
         // depending on the number of arguments, jump into the default argument sequence
         let required_count = u8::try_from(signature.required.len())
-            .map_err(|_| ErrorKind::InternalLimit("parameter count limit exceeded"))?;
+            .map_err(|_| "parameter count limit exceeded")?;
         let default_count = u8::try_from(signature.default.len())
-            .map_err(|_| ErrorKind::InternalLimit("parameter count limit exceeded"))?;
+            .map_err(|_| "parameter count limit exceeded")?;
         
         // "defaults passed" = NArgs - required_count
         self.try_emit_load_local(None, &LocalName::NArgs).unwrap();
@@ -1373,7 +1497,7 @@ impl CodeGenerator<'_> {
             let symbol = Some(param.default.debug_symbol());
             let expr = param.default.variant();
             self.compile_expr(symbol, expr)?;
-            self.emit_instr(None, OpCode::InsertLocal);
+            // self.emit_instr(None, OpCode::InsertLocal);
         }
         
         jump_targets.insert(default_count.into(), self.current_offset());
@@ -1392,7 +1516,7 @@ impl CodeGenerator<'_> {
         debug_assert!(signature.variadic.is_some());
         
         let positional_count = u8::try_from(signature.required.len() + signature.default.len())
-            .map_err(|_| ErrorKind::InternalLimit("parameter count limit exceeded"))?;
+            .map_err(|_| "parameter count limit exceeded")?;
         
         // "variadic count" = NArgs - required_count - default_count
         self.try_emit_load_local(None, &LocalName::NArgs).unwrap();
@@ -1416,7 +1540,7 @@ impl CodeGenerator<'_> {
         
         self.patch_jump_instr(&tuple_jump_site, self.current_offset())?;
         self.emit_instr(None, OpCode::TupleN);
-        self.emit_instr(None, OpCode::InsertLocal);
+        // self.emit_instr(None, OpCode::InsertLocal);
         
         self.patch_jump_instr(&end_jump_site, self.current_offset())?;
         
@@ -1432,19 +1556,19 @@ impl CodeGenerator<'_> {
         let mut required = Vec::new();
         for param in signature.required.iter() {
             let name = self.get_or_make_const(Constant::from(param.name))?;
-            required.push(UnloadedParam { name, decl: param.decl });
+            required.push(UnloadedParam { name, mode: param.mode });
         }
         
         let mut default = Vec::new();
         for param in signature.default.iter() {
             let name = self.get_or_make_const(Constant::from(param.name))?;
-            default.push(UnloadedParam { name, decl: param.decl });
+            default.push(UnloadedParam { name, mode: param.mode });
         }
         
         let mut variadic = None;
         if let Some(param) = &signature.variadic {
             let name = self.get_or_make_const(Constant::from(param.name))?;
-            variadic.replace(UnloadedParam { name, decl: param.decl });
+            variadic.replace(UnloadedParam { name, mode: param.mode });
         }
         
         let signature = UnloadedSignature {
