@@ -15,9 +15,10 @@ pub type UpvalueIndex = u16;
 
                            // width set here so that the longest mnemonic is 16 chars
 const OP_NOP:              u8 = 0x00;
-const OP_EXIT:             u8 = 0x01;
+const OP_EXIT:             u8 = 0x01;  // _ => !
+const OP_ERROR:            u8 = 0x02;  // [ error ] => !
 
-const OP_RETURN:           u8 = 0x02;  // [ ...call frame... ret_value ] => [ ret_value ]
+const OP_RETURN:           u8 = 0x08;  // [ ...call frame... ret_value ] => [ ret_value ]
 
 // The odd argument order is because the VM will swap "nargs" with "arg[n]" after reading "nargs".
 // This is because the function preamble will fill in any missing default arguments such that the
@@ -26,25 +27,28 @@ const OP_RETURN:           u8 = 0x02;  // [ ...call frame... ret_value ] => [ re
 // so swapping it to the beginning is the most efficient way to handle this.
 
 // [ callee arg[n] arg[0] ... arg[n-1] nargs ] => [ ret_value ] 
-const OP_CALL:             u8 = 0x03;
+const OP_CALL:             u8 = 0x09;
 
 // This instruction is needed because when unpacking, the final nargs value needs to be updated dynamically as we iterate.
-// [ callee arg[n] arg[0] ... arg[n-1] nargs arg_seq ] => [ ret_value ]  -- nargs does not include arg_seq! 
-const OP_CALL_UNPACK:      u8 = 0x04;
+// [ callee arg[n] arg[0] ... arg[n-1] nargs arg_unpack ] => [ ret_value ]  -- nargs does not include arg_unpack! 
+const OP_CALL_UNPACK:      u8 = 0x0A;
 
 // 0x10-17        Immediate Values
 
 const OP_POP:              u8 = 0x10;  // [ _ ] => []
 const OP_DROP:             u8 = 0x11;  // (u8); [ value[0] ... value[N] ] => []
 const OP_CLONE:            u8 = 0x12;  // [ value ] => [ value value ]
+const OP_SWAP:             u8 = 0x13;  // (u8); [ value[A] ... value[B] ] => [ value[B] ... value[A] ]
+const OP_SHIFT:            u8 = 0x14;  // (u8); [ value[A] ... value[B] ] => [ ... value[B] value[A] ]
 
-const OP_TUPLE:            u8 = 0x13;  // (u8); [ item[0] ... item[N] ] => [ tuple ]
-const OP_TUPLEN:           u8 = 0x14;  // [ item[0] ... item[N] N ] => [ tuple ]
+const OP_TUPLE:            u8 = 0x18;  // (u8); [ item[0] ... item[N] ] => [ tuple ]
+const OP_TUPLEN:           u8 = 0x19;  // [ item[0] ... item[N] N ] => [ tuple ]
 
 // 0x18-1F        Iteration
 
-const OP_ITER_INIT:        u8 = 0x18;  // [ iter ] => [ iter first_item ]
-const OP_ITER_NEXT:        u8 = 0x19;  // [ iter item ] => [ iter next_item ]
+const OP_ITER_INIT:        u8 = 0x1A;  // [ iterable ] => [ iter state[0] ]
+const OP_ITER_NEXT:        u8 = 0x1B;  // [ iter state[N] ] => [ iter state[N+1] value[N] ]
+const OP_ITER_UNPACK:      u8 = 0x1C;  // [ iter state[N] ] => [ value[N] ... value[M] (M-N) ]
 
 // 0x40-5F        Load/Store
 
@@ -144,6 +148,7 @@ const DBG_DUMP_STRINGS:    u8 = 0xF4;
 pub enum OpCode {
     Nop = OP_NOP,
     Exit = OP_EXIT,
+    Error = OP_ERROR,
     
     Return = OP_RETURN, 
     Call = OP_CALL,
@@ -156,6 +161,10 @@ pub enum OpCode {
     Tuple = OP_TUPLE,
     TupleN = OP_TUPLEN,
     
+    IterInit = OP_ITER_INIT,
+    IterNext = OP_ITER_NEXT,
+    IterUnpack = OP_ITER_UNPACK,
+    
     LoadFunction = OP_LD_FUN,
     LoadFunction16 = OP_LD_FUN_16,
     
@@ -235,6 +244,7 @@ impl OpCode {
         let opcode = match byte {
             OP_NOP => Self::Nop,
             OP_EXIT => Self::Exit,
+            OP_ERROR => Self::Error,
             
             OP_RETURN => Self::Return,
             OP_CALL => Self::Call,
@@ -247,6 +257,10 @@ impl OpCode {
             OP_TUPLE => Self::Tuple,
             OP_TUPLEN => Self::TupleN,
             
+            OP_ITER_INIT => Self::IterInit,
+            OP_ITER_NEXT => Self::IterNext,
+            OP_ITER_UNPACK => Self::IterUnpack,
+            
             OP_LD_FUN => Self::LoadFunction,
             OP_LD_FUN_16 => Self::LoadFunction16,
             
@@ -372,7 +386,7 @@ impl From<OpCode> for u8 {
 
 impl TryFrom<u8> for OpCode {
     type Error = u8;
-    fn try_from(byte: u8) -> Result<Self, Self::Error> {
+    fn try_from(byte: u8) -> Result<Self, u8> {
         if let Some(opcode) = OpCode::from_byte(byte) {
             Ok(opcode)
         } else {
@@ -391,6 +405,7 @@ impl core::fmt::Display for OpCode {
         let mnemonic = match *self {
             Self::Nop => "NOP",
             Self::Exit => "EXIT",
+            Self::Error => "ERROR",
             
             Self::Return => "RETURN",
             Self::Call => "CALL",
@@ -403,8 +418,12 @@ impl core::fmt::Display for OpCode {
             Self::Tuple => "TUPLE",
             Self::TupleN => "TUPLEN",
             
-            Self::LoadFunction => "OP_LD_FUN",
-            Self::LoadFunction16 => "OP_LD_FUN_16",
+            Self::IterInit => "ITER_INIT",
+            Self::IterNext => "ITER_NEXT",
+            Self::IterUnpack => "ITER_UNPACK",
+            
+            Self::LoadFunction => "LD_FUN",
+            Self::LoadFunction16 => "LD_FUN_16",
             
             Self::LoadConst => "LD_CONST",
             Self::LoadConst16 => "LD_CONST_16",
