@@ -2,7 +2,7 @@ use std::collections::VecDeque;
 
 use log::debug;
 
-use crate::language::InternSymbol;
+use crate::language::{InternSymbol, Access};
 use crate::lexer::{TokenMeta, Token, LexerError};
 use crate::runtime::strings::StringInterner;
 use crate::debug::{SourceError, TokenIndex};
@@ -22,7 +22,7 @@ pub use errors::{ParserError, ParseResult};
 use expr::{ExprMeta, Expr, ExprBlock, ConditionalBranch};
 use stmt::{StmtMeta, StmtList, Stmt, Label, ControlFlow};
 use primary::{Primary, Atom, AccessItem};
-use lvalue::{Assignment, LValue, Declaration, DeclType};
+use lvalue::{LValue, AssignType, Assignment};
 use operator::{UnaryOp, BinaryOp, Precedence, PRECEDENCE_START, PRECEDENCE_END};
 use fundefs::{FunctionDef, SignatureDef, ParamDef, DefaultDef};
 use errors::{ErrorKind, ErrorContext, ContextTag};
@@ -37,7 +37,7 @@ pub struct Parser<'h, T> where T: Iterator<Item=Result<TokenMeta, LexerError>> {
     errors: VecDeque<ParserError>,
 }
 
-impl<'h, T> Iterator for Parser<'h, T> where T: Iterator<Item=Result<TokenMeta, LexerError>> {
+impl<T> Iterator for Parser<'_, T> where T: Iterator<Item=Result<TokenMeta, LexerError>> {
     type Item = Result<StmtMeta, ParserError>;
     fn next(&mut self) -> Option<Self::Item> { self.next_stmt() }
 }
@@ -51,7 +51,9 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             errors: VecDeque::new(),
         }
     }
+}
     
+impl<I> Parser<'_, I> where I: Iterator<Item=Result<TokenMeta, LexerError>> {
     /// for debugging
     fn current_index(&mut self) -> TokenIndex { 
         let token = self.peek().unwrap();
@@ -92,7 +94,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
     fn intern_str(&mut self, string: impl AsRef<str>) -> InternSymbol {
         self.interner.get_or_intern(string)
     }
-    
+
     pub fn next_stmt(&mut self) -> Option<Result<StmtMeta, ParserError>> {
         let mut ctx = ErrorContext::new(ContextTag::TopLevel);
         
@@ -213,8 +215,8 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         
         debug!("done.");
     }
-    
-    /*** Statement Parsing ***/
+
+    /* Statement Parsing */
     
     fn parse_stmt(&mut self, ctx: &mut ErrorContext) -> ParseResult<StmtMeta> {
         // skip statement separators
@@ -261,10 +263,17 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                 };
                 
                 let message = format!("\"{}\" is not allowed here", name);
-                return Err(ErrorKind::SyntaxError(message).into());
-            },
+                return Err(message.as_str().into());
+            }
             
-            _ => Stmt::Expression(self.parse_expr_variant(ctx)?),
+            // expression statements
+            _ => match self.parse_expr_variant(ctx)? {
+                
+                Expr::Ellipsis(..) => 
+                    return Err("\"...\" is not allowed here".into()),
+                
+                expr => Stmt::Expression(expr)
+            }
         };
         Ok(stmt)
     }
@@ -310,7 +319,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         ctx.set_end(&next);
         
         if !matches!(next.token, Token::Do) {
-            return Err("missing \"do\" after condition in while-loop".into());
+            return Err("expected \"do\" after condition in while-loop".into());
         }
         
         let body = self.parse_stmt_list(ctx, |token| matches!(token, Token::End))?;
@@ -456,8 +465,8 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         ctx.pop_extend();
         Ok(Some(control_flow))
     }
-    
-    /*** Expression Parsing ***/
+
+    /* Expression Parsing */
     
     fn parse_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<ExprMeta> {
         ctx.push(ContextTag::ExprMeta);
@@ -471,10 +480,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
     
     // the top of the recursive descent stack for expressions
     fn parse_expr_variant(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
-        match self.peek()?.token {
-            Token::Var | Token::Let => self.parse_declaration_expr(ctx),
-            _ => self.parse_assignment_expr(ctx)
-        }
+        self.parse_assignment_expr(ctx)
     }
     
     /*
@@ -493,10 +499,10 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
     
     fn parse_assignment_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
         
-        let optional_token = 
-            if let Token::NonLocal = self.peek()?.token { Some(self.advance().unwrap()) }
-            else { None };
+        // check for lvalue modifier
+        let assign = self.try_parse_assign_keyword(ctx)?;
         
+        // parse LHS
         let expr = self.parse_tuple_expr(ctx)?;
         
         let next = self.peek()?;
@@ -505,81 +511,38 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             ctx.push_continuation(ContextTag::AssignmentExpr, None);
             ctx.set_end(&self.advance().unwrap());
             
-            if let Some(token) = optional_token.as_ref() {
-                ctx.set_start(token);
-            }
-            
-            // LHS of assignment has to be an lvalue
-            let lhs = LValue::try_from(expr).map_err(|_| ParserError::from("can't assign to this"))?;
-            
-            // update-assignment not allowed with tuples
-            if op.is_some() && matches!(lhs, LValue::Tuple(..)) {
-                return Err("can't combine tuple-assignment with update-assignment".into());
-            }
+            // LHS of assignment must be an lvalue
+            let lhs = LValue::try_from(expr)
+                .map_err(|_| ParserError::from("can't assign to this"))?;
             
+            // Parse RHS
             let rhs = self.parse_expr_variant(ctx)?;
             
             ctx.pop_extend();
             
-            let nonlocal = matches!(optional_token.map(|tok| tok.token), Some(Token::NonLocal));
+            let assign = Assignment {
+                lhs, op, rhs,
+                assign: assign.unwrap_or(AssignType::AssignLocal),
+            };
             
-            let assign = Box::new(Assignment { lhs, op, rhs, nonlocal });
-            return Ok(Expr::Assignment(assign));
+            Ok(Expr::Assignment(Box::new(assign)))
             
-        } else if optional_token.is_some() {
-            return Err("expected an assignment expression after \"nonlocal\"".into())
+        } else if let Some(assign) = assign {
+            let assign = match assign {
+                AssignType::AssignLocal => "local",
+                AssignType::AssignNonLocal => "nonlocal",
+                AssignType::DeclImmutable => "let",
+                AssignType::DeclMutable => "var",
+            };
+            let message = format!("expected an assignment expression after \"{}\"", assign);
+            
+            Err(message.as_str().into())
+            
+        } else {
+            
+            Ok(expr)
         }
         
-        Ok(expr)
-    }
-    
-    /*
-        declaration_expression ::= ( "let" | "var" ) lvalue_expr_annotated "=" expression ;
-    */
-    fn parse_declaration_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
-        let next = self.advance()?;
-        
-        ctx.push(ContextTag::DeclarationExpr);
-        ctx.set_start(&next);
-        
-        let decl = match next.token {
-            Token::Let => DeclType::Immutable,
-            Token::Var => DeclType::Mutable,
-            _ => panic!("invalid decl token"),
-        };
-        
-        let expr = self.parse_tuple_expr(ctx)?;
-        let lhs = LValue::try_from(expr).map_err(|_| ParserError::from("can't assign to this"))?;
-        if !Self::is_lvalue_valid_for_decl(&lhs) {
-            return Err("only names can be declared as variables".into());
-        }
-        
-        // check for and consume "="
-        let next = self.advance()?;
-        ctx.set_end(&next);
-        
-        match Self::which_assignment_op(&next.token) {
-            None => return Err("missing \"=\" in variable declaration".into()),
-            Some(op) => if op.is_some() {
-                return Err("update-assignment is not allowed in a variable declaration".into());
-            }
-        }
-        
-        let init = self.parse_expr_variant(ctx)?;
-        
-        ctx.pop_extend();
-        
-        let decl = Box::new(Declaration { decl, lhs, init });
-        Ok(Expr::Declaration(decl))
-    }
-    
-    fn is_lvalue_valid_for_decl(lvalue: &LValue) -> bool {
-        match lvalue {
-            LValue::Identifier(..) => true,
-            LValue::Tuple(lvalue_list) 
-                => lvalue_list.iter().all(|lvalue| Self::is_lvalue_valid_for_decl(lvalue)),
-            _ => false,
-        }
     }
     
     fn parse_tuple_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
@@ -595,6 +558,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         let mut tuple_exprs = Vec::new();
         loop {
             let next = self.peek()?;
+            
             if !matches!(next.token, Token::Comma) {
                 break;
             }
@@ -632,6 +596,8 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         }
     }
     
+
+
     /*
         Binary operator syntax:
         
@@ -781,11 +747,35 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                 Expr::Echo(Box::new(self.parse_expr_variant(ctx)?))
             },
             
-            _ => self.parse_primary(ctx)?,
+            _ => self.parse_unpack_expr(ctx)?,
         };
         Ok(expr)
     }
     
+    fn parse_unpack_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
+        // check for plain ellipsis
+        let next = self.peek()?;
+        if matches!(next.token, Token::Ellipsis) {
+            return Ok(Expr::Ellipsis(None));
+        }
+        
+        let expr = self.parse_primary(ctx)?;
+        
+        let next = self.peek()?;
+        if matches!(next.token, Token::Ellipsis) {
+            ctx.set_end(&self.advance().unwrap());
+            
+            // Having multiple "..."s next to each other is really bad for readability
+            // So require that they are separated by parens, e.g. (((foo...)...)...)
+            if matches!(expr, Expr::Ellipsis(..)) {
+                return Err("nested use of \"...\" must be enclosed in parentheses".into());
+            }
+            return Ok(Expr::Ellipsis(Some(Box::new(expr))));
+        }
+        
+        Ok(expr)
+    }
+    
     fn parse_expr_label(&mut self, ctx: &mut ErrorContext) -> ParseResult<Expr> {
         let label = self.try_parse_label(ctx)?.unwrap();
         
@@ -836,7 +826,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             ctx.set_end(&next);
             
             if !matches!(next.token, Token::Then) {
-                return Err("missing \"then\" after condition in if-expression".into());
+                return Err("expected \"then\" after condition in if-expression".into());
             }
             
             let stmt_list = self.parse_stmt_list(ctx, |token| matches!(token, Token::Elif | Token::Else | Token::End))?;
@@ -896,13 +886,15 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                 function_def.signature.name.replace(*name);
             }
             
-            let function_decl = Declaration {
-                decl: DeclType::Immutable,
+            let fun_decl = Assignment {
+                assign: AssignType::DeclImmutable,
+                op: None,
                 lhs: lvalue,
-                init: Expr::FunctionDef(function_def),
+                rhs: Expr::FunctionDef(function_def),
             };
+            
+            Ok(Expr::Assignment(Box::new(fun_decl)))
 
-            Ok(Expr::Declaration(Box::new(function_decl)))
         } else {
 
             Ok(Expr::FunctionDef(function_def))
@@ -948,14 +940,14 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         // function parameter list
         
         if !matches!(next.token, Token::OpenParen) {
-            return Err("missing opening \"(\" before parameter list".into());
+            return Err("expected opening \"(\" before parameter list".into());
         }
         
         let signature = self.parse_function_param_list(ctx)?;
         
         let next = self.advance()?;
         if !matches!(next.token, Token::CloseParen) {
-            return Err("missing closing \")\" after parameter list".into());
+            return Err("expected closing \")\" after parameter list".into());
         }
         
         // function body
@@ -989,19 +981,17 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             
             // mutability modifier
             
-            let decl = match next.token {
-                Token::Var => Some(DeclType::Mutable),
-                Token::Let => Some(DeclType::Immutable),
+            let mode = match next.token {
+                Token::Var => Some(Access::ReadWrite),
+                Token::Let => Some(Access::ReadOnly),
                 Token::Identifier(..) => None,
                 _ => return Err("invalid parameter".into()),
             };
             
-            if decl.is_some() {
+            if mode.is_some() {
                 ctx.set_end(&self.advance().unwrap());
             }
-            
-            let decl = decl.unwrap_or(DeclType::Immutable);
-            
+                        
             // name
             
             let next = self.advance()?;
@@ -1039,6 +1029,8 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             
             // expect either a comma "," or the closing ")"
             let next = self.peek()?;
+
+            let mode = mode.unwrap_or(Access::ReadOnly);
             match next.token {
                 // variadic parameter
                 Token::Comma if is_variadic => {
@@ -1047,18 +1039,18 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                 
                 Token::CloseParen if is_variadic => {
                     debug_assert!(default_value.is_none());
-                    variadic.replace(ParamDef { name, decl });
+                    variadic.replace(ParamDef { name, mode });
                 },
                 
                 // normal parameter
                 Token::Comma | Token::CloseParen if !is_variadic => {
                     if let Some(default_expr) = default_value {
-                        default.push(DefaultDef { name, decl, default: default_expr });
+                        default.push(DefaultDef { name, mode, default: default_expr });
                     } else {
                         if !default.is_empty() {
                             return Err("cannot have a non-default parameter after a default parameter".into());
                         }
-                        required.push(ParamDef { name, decl });
+                        required.push(ParamDef { name, mode });
                     }
                 },
                 
@@ -1134,11 +1126,9 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                 
                 // subscript ::= "[" expression "]" ;
                 Token::OpenSquare => items.push(self.parse_index_access(ctx)?),
-                
-                
+                                
                 // invocation ::= "(" ")" | "(" argument ( "," argument )* ")" ; 
                 // argument ::= expression ( "..." )? ;  (* "..." is for argument unpacking syntax *)
-                
                 // invocations are not allowed to be on a separate line from the invocation receiver
                 Token::OpenParen if !next.newline => items.push(self.parse_invocation(ctx)?),
                 
@@ -1211,7 +1201,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         ctx.set_end(&next);
         
         if !matches!(next.token, Token::CloseSquare) {
-            return Err("missing closing \"]\"".into());
+            return Err("expected closing \"]\"".into());
         }
         
         ctx.pop_extend();
@@ -1226,8 +1216,6 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
         debug_assert!(matches!(next.token, Token::OpenParen));
         
         let mut args = Vec::new();
-        let mut unpack = None;
-        
         
         // check for empty argument list
         if matches!(self.peek()?.token, Token::CloseParen) {
@@ -1242,35 +1230,21 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             
             let (expr, symbol) = self.parse_expr(ctx)?.take();
             
-            if let Expr::Tuple(arg_list) = expr {
-                args.extend(arg_list.into_vec().into_iter());
+            if let Expr::Tuple(items) = expr {
+                args.extend(items.into_vec().into_iter());
             } else {
                 args.push(ExprMeta::new(expr, symbol));
             }
             
-            // check for unpack syntax
-            if matches!(self.peek()?.token, Token::Ellipsis) {
-                ctx.set_end(&self.advance().unwrap());
-                
-                if let Some(expr) = args.pop() {
-                    unpack.replace(expr);
-                } else {
-                    return Err("no argument to unpack".into());
-                }
-            }
-            
             // check for close paren
             let next = self.advance()?;
             ctx.set_end(&next);
             if !matches!(next.token, Token::CloseParen) {
-                return Err("missing closing \")\" after argument list".into());
+                return Err("expected closing \")\" after argument list".into());
             }
         }
         
-        let invocation = AccessItem::Invoke {
-            args: args.into_boxed_slice(),
-            unpack,
-        };
+        let invocation = AccessItem::Invoke(args.into_boxed_slice());
         
         ctx.pop_extend();
         Ok(invocation)
@@ -1305,13 +1279,15 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
                     Atom::StringLiteral(self.intern_str(value))
                 },
                 
-                
                 // Error productions
                 Token::Class | Token::Fun | Token::If | Token::Var | Token::Let | Token::Begin | Token::Label(..) => {
                     let name = match next.token {
                         Token::Class => "class definitions",
                         Token::Fun => "function definitions",
-                        Token::Let | Token::Var => "variable declarations",
+                        Token::Let => "\"let\"",
+                        Token::Var => "\"var\"",
+                        Token::Local => "\"local\"",
+                        Token::NonLocal => "\"nonlocal\"",
                         Token::Begin => "block expressions",
                         _ => "this expression",
                     };
@@ -1331,7 +1307,7 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             Ok(atom)
         }
     }
-    
+
     fn parse_group_expr(&mut self, ctx: &mut ErrorContext) -> ParseResult<Atom> {
         ctx.push(ContextTag::Group);
         
@@ -1346,18 +1322,46 @@ impl<'h, I> Parser<'h, I> where I: Iterator<Item=Result<TokenMeta, LexerError>>
             ctx.pop_extend();
             return Ok(Atom::EmptyTuple);
         }
-        
+
+        // Check for lvalue modifier
+        let modifier = self.try_parse_assign_keyword(ctx)?;
+
+        // Parse inner expression
         let expr = self.parse_expr_variant(ctx)?;
         
         // Consume and check closing paren
         let next = self.advance()?;
         ctx.set_end(&next);
         if !matches!(next.token, Token::CloseParen) {
-            return Err("missing closing \")\"".into());
+            return Err("expected closing \")\"".into());
         }
         
         ctx.pop_extend();
-        Ok(Atom::Group(Box::new(expr)))
+        Ok(Atom::Group {
+            modifier, inner: Box::new(expr),
+        })
     }
-    
+
+    /* LValue Parsing */
+    fn try_parse_assign_keyword(&mut self, ctx: &mut ErrorContext) -> ParseResult<Option<AssignType>> {
+        let next = self.peek()?;
+
+        // if we see any of these, then this is definitely an LValue
+        let modifier = match next.token {
+            Token::Let => Some(AssignType::DeclImmutable),
+            Token::Var => Some(AssignType::DeclMutable),
+            Token::Local => Some(AssignType::AssignLocal),
+            Token::NonLocal => Some(AssignType::AssignNonLocal),
+            _ => None,
+        };
+
+        if modifier.is_some() {
+            ctx.push(ContextTag::LValue);
+            ctx.set_start(&self.advance().unwrap());
+            ctx.pop_extend();
+            return Ok(modifier);
+        }
+        Ok(None)
+    }
+
 }
