@@ -35,11 +35,11 @@ pub enum ContextTag {
     WhileLoop,
     ExprMeta,
     ExprList,
+    Expr,
     BlockExpr,
     IfExpr,
     FunDefExpr,
     FunParam,
-    DeclarationExpr,
     AssignmentExpr,
     BinaryOpExpr,
     UnaryOpExpr,
@@ -51,6 +51,7 @@ pub enum ContextTag {
     TupleCtor,
     Atom,
     Group,
+    LValue,
     Label,
 }
 
