@@ -1,8 +1,8 @@
 use core::fmt;
-use crate::parser::lvalue::DeclType;
+use crate::language::Access;
 use crate::runtime::Variant;
 use crate::runtime::strings::{StringValue, StringSymbol, StrBuffer, STRING_TABLE};
-use crate::runtime::errors::{ExecResult, ErrorKind};
+use crate::runtime::errors::{ExecResult, RuntimeError};
 
 
 #[derive(Clone, Debug)]
@@ -71,17 +71,11 @@ impl Signature {
     
     pub fn check_args(&self, args: &[Variant]) -> ExecResult<()> {
         if args.len() < self.required.len() {
-            return Err(ErrorKind::MissingArguments { 
-                signature: Box::new(self.clone()),
-                nargs: args.len() 
-            }.into())
+            return Err(RuntimeError::missing_arguments(self, args.len()))
         }
         
         if matches!(self.max_arity(), Some(max_arity) if args.len() > max_arity) {
-            return Err(ErrorKind::TooManyArguments {
-                signature: Box::new(self.clone()),
-                nargs: args.len()
-            }.into())
+            return Err(RuntimeError::too_many_arguments(self, args.len()))
         }
         
         Ok(())
@@ -138,24 +132,16 @@ pub struct BoundArgs<'a> {
 #[derive(Clone, Debug)]
 pub struct Parameter {
     name: StringSymbol,
-    decl: DeclType,
+    mode: Access,
 }
 
 impl Parameter {
-    pub fn new(name: impl Into<StringSymbol>, decl: DeclType) -> Self {
-        Self { name: name.into(), decl }
-    }
-    
-    pub fn new_let(name: impl Into<StringSymbol>) -> Self {
-        Self::new(name, DeclType::Immutable)
-    }
-    
-    pub fn new_var(name: impl Into<StringSymbol>) -> Self {
-        Self::new(name, DeclType::Mutable)
+    pub fn new(name: impl Into<StringSymbol>, mode: Access) -> Self {
+        Self { name: name.into(), mode }
     }
     
     pub fn name(&self) -> &StringSymbol { &self.name }
-    pub fn decl(&self) -> &DeclType { &self.decl }
+    pub fn mode(&self) -> &Access { &self.mode }
 }
 
 
