@@ -2,8 +2,8 @@ use core::fmt::{self, Write};
 use crate::runtime::Variant;
 use crate::runtime::gc::{Gc, GcTrace};
 use crate::runtime::strings::{StringValue, static_symbol};
-use crate::runtime::types::{Type, MetaObject};
-use crate::runtime::errors::{ExecResult, ErrorKind};
+use crate::runtime::types::{Type, MetaObject, IterState, UserIterator};
+use crate::runtime::errors::{ExecResult, RuntimeError};
 
 #[derive(Clone, Copy)]
 pub enum Tuple {
@@ -11,18 +11,6 @@ pub enum Tuple {
     NonEmpty(Gc<[Variant]>),
 }
 
-unsafe impl GcTrace for Box<[Variant]> {
-    fn trace(&self) {
-        for item in self.iter() {
-            item.trace();
-        }
-    }
-    
-    fn size_hint(&self) -> usize {
-        core::mem::size_of::<Variant>() * self.len()
-    }
-}
-
 impl Default for Tuple {
     fn default() -> Self { Self::Empty }
 }
@@ -75,6 +63,12 @@ impl MetaObject for Tuple {
         Some(Ok(Tuple::len(self)))
     }
     
+    fn iter_init(&self) -> Option<ExecResult<IterState>> {
+        let iter: Box<dyn UserIterator> = Box::new(TupleIter(*self));
+        let iter = Gc::from_box(iter);
+        iter.iter_init()
+    }
+    
     fn fmt_echo(&self) -> ExecResult<StringValue> {
         match self {
             Self::Empty => Ok(StringValue::from(static_symbol!("()"))),
@@ -85,11 +79,11 @@ impl MetaObject for Tuple {
                 let mut buf = String::new();
                 
                 write!(&mut buf, "({}", first.fmt_echo()?)
-                    .map_err(ErrorKind::from)?;
+                    .map_err(|err| RuntimeError::other(err.to_string()))?;
                 
                 for item in rest.iter() {
                     write!(&mut buf, ", {}", item.fmt_echo()?)
-                        .map_err(ErrorKind::from)?;
+                        .map_err(|err| RuntimeError::other(err.to_string()))?;
                 }
                 buf.push(')');
 
@@ -109,3 +103,41 @@ impl fmt::Debug for Tuple {
         tuple.finish()
     }
 }
+
+// Tuple Iterator
+#[derive(Debug)]
+struct TupleIter(Tuple);
+
+unsafe impl GcTrace for TupleIter {
+    fn trace(&self) {
+        self.0.trace()
+    }
+}
+
+impl UserIterator for TupleIter {
+    fn get_item(&self, state: &Variant) -> ExecResult<Variant> {
+        let idx = usize::try_from(state.as_int()?)
+            .map_err(|_| RuntimeError::invalid_value("invalid state"))?;
+        
+        let items = self.0.as_ref();
+        Ok(items[idx])
+    }
+    
+    fn next_state(&self, state: Option<&Variant>) -> ExecResult<Variant> {
+        let next = match state {
+            Some(state) => state.as_int()?
+                .checked_add(1)
+                .ok_or(RuntimeError::overflow_error())?,
+            
+            None => 0,
+        };
+        
+        let next_idx = usize::try_from(next)
+            .map_err(|_| RuntimeError::invalid_value("invalid state"))?;
+        
+        if next_idx >= self.0.len() {
+            return Ok(Variant::Nil)
+        }
+        Ok(Variant::from(next))
+    }
+}
\ No newline at end of file
